@@ -5,8 +5,8 @@
 ;; Author: William Bosch-Bello <williamsbosch@gmail.com>
 ;; Maintainer: William Bosch-Bello <williamsbosch@gmail.com>
 ;; Created: August 16, 2025
-;; Package-Version: 1.4.1
-;; Package-Revision: v1.4.1-0-g9c246fe5c716
+;; Package-Version: 1.6.4
+;; Package-Revision: v1.6.4-0-gf5dd83600fa1
 ;; Package-Requires: ((emacs "29.1") (async "1.9") (promise "1.1"))
 ;; Keywords: aws, iam, org, babel, tools
 ;; URL: https://github.com/will-abb/org-aws-iam-role
@@ -38,17 +38,18 @@
 ;; - Interactive browsing and selection of IAM roles.
 ;; - Full display of all policy types: Trust Policy, Permissions Boundary,
 ;;   AWS-Managed, Customer-Managed, and Inline policies.
-;; - Direct modification of any policy through Org Babel source blocks. Use
-;;   header arguments like `:create`, `:delete`, or `:detach` for full
-;;   CRUD (Create, Read, Update, Delete) operations.
+;; - Direct modification via Org Babel. Supports "Upsert" logic: automatically
+;;   creates policies if they don't exist, or updates them if they do.
+;; - Smart Version Management: Automatically offers to delete the oldest
+;;   policy version if the AWS version limit is reached during an update.
+;; - Advanced Header Arguments: Support for `:tags`, `:path`, and sequential
+;;   `:detach` + `:delete` operations.
 ;; - Built-in IAM Policy Simulator to test a role's permissions against
 ;;   specific AWS actions and resources.
-;; - View a combined JSON policy of all permission policies for easy export
-;;   or analysis.
+;; - View a combined JSON policy of all permission policies for easy export.
 ;; - Asynchronous fetching of initial role and policy data for a fast UI.
 ;; - Safe by default: the role viewer buffer opens in read-only mode.
 ;; - Ability to easily switch between different AWS profiles.
-;; - Clear feedback on command success or failure in Babel results blocks.
 
 ;; Keybindings:
 ;;
@@ -57,6 +58,7 @@
 ;; - C-c C-s: Simulate the role's policies against specific actions.
 ;; - C-c C-j: View a combined JSON of all permission policies.
 ;; - C-c C-a: Get service last accessed details for the role.
+;; - C-c C-m: Find the last modified date for the role or its policies.
 ;; - C-c C-c: Inside a source block, apply changes to AWS.
 ;; - C-c (:   Hide all property drawers.
 ;; - C-c ):   Reveal all property drawers.
@@ -123,6 +125,14 @@ If ROLE-NAME is provided programmatically, skip prompting."
           (completing-read "Select AWS profile: " profiles nil t))
     (message "Set IAM Role AWS profile to: %s" org-aws-iam-role-profile)))
 
+(defun org-aws-iam-role--get-account-id ()
+  "Fetch the current AWS Account ID using sts get-caller-identity."
+  (let* ((cmd (format "aws sts get-caller-identity --output json%s"
+                      (org-aws-iam-role--cli-profile-arg)))
+         (json (shell-command-to-string cmd))
+         (parsed (json-parse-string json :object-type 'alist)))
+    (alist-get 'Account parsed)))
+
 (defun org-aws-iam-role-toggle-read-only ()
   "Toggle read-only mode in the current buffer and provide feedback."
   (interactive)
@@ -153,7 +163,6 @@ If ROLE-NAME is provided programmatically, skip prompting."
 
 (defun org-aws-iam-role--format-tags (tags)
   "Format AWS TAGS from a list of alists into a single JSON string.
-
 Argument TAGS is a list of alists of the form
 \(\(\"Key\" . \"k1\"\) \(\"Value\" . \"v1\"\)\)."
   (when tags
@@ -198,7 +207,6 @@ Argument TAGS is a list of alists of the form
 
 (defun org-aws-iam-role--policy-get-metadata-async (policy-arn)
   "Fetch policy metadata JSON asynchronously for POLICY-ARN.
-
 Returns a promise that resolves with the raw JSON string from the
 `get-policy` command."
   (let* ((cmd (format "aws iam get-policy --policy-arn %s --output json%s"
@@ -209,7 +217,6 @@ Returns a promise that resolves with the raw JSON string from the
 
 (defun org-aws-iam-role--policy-get-version-document-async (policy-arn version-id)
   "Fetch policy document JSON for POLICY-ARN and VERSION-ID.
-
 This is an asynchronous operation using `get-policy-version`.
 Returns a promise that resolves with the raw JSON string."
   (let* ((cmd (format "aws iam get-policy-version --policy-arn %s --version-id %s --output json%s"
@@ -221,7 +228,6 @@ Returns a promise that resolves with the raw JSON string."
 
 (defun org-aws-iam-role-policy--construct-from-data (metadata policy-type document-json)
   "Construct an `org-aws-iam-role-policy` struct from resolved data.
-
 METADATA is the parsed `Policy` alist from `get-policy`.
 POLICY-TYPE is the type symbol (e.g., `aws-managed`).
 DOCUMENT-JSON is the raw JSON string from `get-policy-version`."
@@ -249,7 +255,6 @@ DOCUMENT-JSON is the raw JSON string from `get-policy-version`."
 
 (defun org-aws-iam-role--policy-from-arn-async (policy-arn policy-type)
   "Create an `org-aws-iam-role-policy` struct asynchronously from a policy ARN.
-
 Argument POLICY-ARN is the ARN of the IAM policy.
 Argument POLICY-TYPE is the type of the IAM policy.
 Returns a promise that resolves with the complete
@@ -281,7 +286,6 @@ Returns a promise that resolves with the complete
 
 (defun org-aws-iam-role-inline-policy--construct-from-json (policy-name json)
   "Construct an inline `org-aws-iam-role-policy` struct from its JSON.
-
 POLICY-NAME is the name of the inline policy. JSON is the raw
 string from the `get-role-policy` AWS CLI command."
   (let* ((parsed (json-parse-string json :object-type 'alist :array-type 'list))
@@ -298,7 +302,6 @@ string from the `get-role-policy` AWS CLI command."
 
 (defun org-aws-iam-role--inline-policy-from-name-async (role-name policy-name)
   "Fetch an inline policy asynchronously and construct a struct.
-
 Argument ROLE-NAME is the name of the IAM role.
 Argument POLICY-NAME is the name of the inline policy.
 Returns a promise that resolves with the `org-aws-iam-role-policy` struct."
@@ -316,7 +319,6 @@ Returns a promise that resolves with the `org-aws-iam-role-policy` struct."
 
 (defun org-aws-iam-role--fetch-roles-page (marker)
   "Fetch a single page of IAM roles from AWS.
-
 If MARKER is non-nil, it's used as the `--starting-token`.
 Returns a cons cell: (LIST-OF-ROLES . NEXT-MARKER)."
   (let* ((cmd (format "aws iam list-roles --output json%s%s"
@@ -331,17 +333,12 @@ Returns a cons cell: (LIST-OF-ROLES . NEXT-MARKER)."
 (defun org-aws-iam-role--list-names ()
   "Return a list of all IAM role names, handling pagination."
   (let ((all-roles '())
-        (marker nil)
-        (first-run t))
-    ;; Loop until the AWS API returns no more pages.
-    ;; The `first-run` flag ensures the loop runs at least once when marker starts as nil.
-    (while (or first-run marker)
-      (let* ((page-result (org-aws-iam-role--fetch-roles-page marker))
-             (roles-on-page (car page-result))
-             (next-marker (cdr page-result)))
-        (setq all-roles (nconc all-roles roles-on-page))
-        (setq marker next-marker)
-        (setq first-run nil)))
+        (marker nil))
+    (cl-loop do
+             (let* ((page-result (org-aws-iam-role--fetch-roles-page marker)))
+               (setq all-roles (nconc all-roles (car page-result)))
+               (setq marker (cdr page-result)))
+             while marker)
     (mapcar (lambda (r) (alist-get 'RoleName r)) all-roles)))
 
 (defun org-aws-iam-role--get-full (role-name)
@@ -355,7 +352,6 @@ Returns a cons cell: (LIST-OF-ROLES . NEXT-MARKER)."
 
 (defun org-aws-iam-role--construct (obj)
   "Create an `org-aws-iam-role` struct from a full `get-role` object.
-
 Argument OBJ is the JSON object returned by `get-role`."
   ;; PermissionsBoundary and RoleLastUsed can be nil, so we get them first.
   (let ((pb (alist-get 'PermissionsBoundary obj))
@@ -395,7 +391,6 @@ Argument OBJ is the JSON object returned by `get-role`."
 
 (defun org-aws-iam-role--split-managed-policies (attached)
   "Split ATTACHED managed policies into (customer . aws) buckets.
-
 Each bucket keeps the full alist for each policy item."
   (let ((customer '()) (aws '()))
     (dolist (p attached)
@@ -463,6 +458,7 @@ Each bucket keeps the full alist for each policy item."
     (insert (format ":Updated: %s\n" (or (org-aws-iam-role-policy-update-date policy) "nil")))
     (insert (format ":AttachmentCount: %s\n" (or (org-aws-iam-role-policy-attachment-count policy) "nil")))
     (insert (format ":DefaultVersion: %s\n" (or (org-aws-iam-role-policy-default-version-id policy) "nil")))
+    (insert (format ":Tags: %s\n" (or (org-aws-iam-role--format-tags (org-aws-iam-role-policy-tags policy)) "nil")))
     (insert ":END:\n")
     (insert (format "#+BEGIN_SRC aws-iam :role-name \"%s\" :policy-name \"%s\" :policy-type \"%s\" :arn \"%s\" :results output\n"
                     role-name
@@ -498,7 +494,6 @@ INLINE-POLICY-NAMES is a list of inline policy names."
 
 (defun org-aws-iam-role--get-all-policies-async (role)
   "Fetch all attached, inline, and boundary policies for ROLE.
-
 This function is asynchronous and returns a single promise that
 resolves with a vector of `org-aws-iam-role-policy` structs when
 all underlying fetches are complete. Returns nil if no policies
@@ -514,7 +509,6 @@ are found."
 
 (defun org-aws-iam-role--insert-policies-section (all-policies-vector boundary-arn role-name)
   "Render a vector of fetched policies into the current buffer.
-
 ALL-POLICIES-VECTOR is the result from a `promise-all' call.
 BOUNDARY-ARN is the original ARN of the boundary policy.
 ROLE-NAME is the name of the parent IAM role."
@@ -544,17 +538,20 @@ ROLE-NAME is the name of the parent IAM role."
   (insert "** Applying Changes via Babel\n")
   (insert "All actions are performed by executing an =aws-iam= source block with =C-c C-c=.\n")
   (insert "You will be asked to confirm before any change is applied.\n")
-  (insert "\n- *To Create or Update an Inline Policy*, simply write or edit its source block and execute it. No special flag is needed.\n")
-  (insert "- *To Update any other policy*, edit its source block and execute.\n")
-  (insert "\nUse header arguments for other actions:\n")
-  (insert "- =:create t=     :: Creates a new *customer-managed* policy.\n")
-  (insert "- =:delete t=     :: Deletes a policy. For managed policies, this will fail if the policy is still attached to any role, user, or group.\n")
-  (insert "- =:detach t=     :: Detaches a *managed* policy from the current role.\n")
+  (insert "\n- *To Create a New Policy*: Provide a unique =:policy-name= and optional =:path=. If the ARN does not exist, it will be created.\n")
+  (insert "- *To Update a Policy*: Edit the JSON. If the policy has reached the version limit, you will be offered to delete the oldest version.\n")
+  (insert "- *To Tag*: Add the =:tags= header. Tagging is applied /before/ updates.\n")
+  (insert "\nUse header arguments for specific actions:\n")
+  (insert "- =:delete t=     :: Deletes the policy. If versions exist, you will be asked to recursively delete them.\n")
+  (insert "- =:detach t=     :: Detaches the policy from the current role. Can be combined with =:delete t=.\n")
+  (insert "- =:tags=         :: A string of tags (e.g. \"Key=Owner,Value=Dev Key=Stage,Value=Prod\").\n")
+  (insert "- =:path=         :: The IAM path for creation (e.g. \"/service-role/\"). Defaults to \"/\".\n")
   (insert "\n** Keybindings\n")
   (insert "- =C-c C-e= :: Toggle read-only mode to allow/prevent edits.\n")
   (insert "- =C-c C-s= :: Simulate the role's policies against specific actions.\n")
   (insert "- =C-c C-j= :: View a combined JSON of all permission policies.\n")
   (insert "- =C-c C-a= :: Get service last accessed details for the role.\n")
+  (insert "- =C-c C-m= :: Find the last modified date for the role or its policies.\n")
   (insert "- =C-c C-c= :: Inside a source block, apply changes to AWS.\n")
   (insert "- =C-c (= :: Hide all property drawers.\n")
   (insert "- =C-c )= :: Reveal all property drawers.\n\n"))
@@ -563,7 +560,7 @@ ROLE-NAME is the name of the parent IAM role."
   "Callback to populate BUF with fetched policies for ROLE.
 ALL-POLICIES-VECTOR is the resolved vector of policy structs."
   ;; Catch any error during rendering to prevent the callback from crashing.
-  (condition-case nil
+  (condition-case err
       (with-current-buffer buf
         (let ((boundary-arn (org-aws-iam-role-permissions-boundary-arn role))
               (role-name (org-aws-iam-role-name role)))
@@ -571,7 +568,7 @@ ALL-POLICIES-VECTOR is the resolved vector of policy structs."
           (org-aws-iam-role--insert-policies-section all-policies-vector boundary-arn role-name)
           ;; Insert remaining synchronous sections and finalize the buffer.
           (org-aws-iam-role--insert-remaining-sections-and-finalize role buf)))
-    (error nil)))
+    (error (message "Error populating buffer: %s" (error-message-string err)))))
 
 (defun org-aws-iam-role--populate-role-buffer (role buf)
   "Insert all details for ROLE and its policies into the buffer BUF.
@@ -605,6 +602,7 @@ information."
     (local-set-key (kbd "C-c C-s") #'org-aws-iam-role-simulate-from-buffer)
     (local-set-key (kbd "C-c C-j") #'org-aws-iam-role-combine-permissions-from-buffer)
     (local-set-key (kbd "C-c C-a") #'org-aws-iam-role-get-last-accessed)
+    (local-set-key (kbd "C-c C-m") #'org-aws-iam-role-get-last-modified)
     (local-set-key (kbd "C-c (") #'org-fold-hide-drawer-all)
     (local-set-key (kbd "C-c )") #'org-fold-show-all)
     (goto-char (point-min))    (when org-aws-iam-role-read-only-by-default
@@ -865,15 +863,77 @@ POLICY-DOCUMENT is the policy JSON string."
       (or (equal val-str "t")
           (equal val-str "true")))))
 
-(defun org-aws-iam-role--babel-cmd-create-policy (policy-name policy-document)
+(defun org-aws-iam-role--policy-exists-p (policy-arn)
+  "Return t if the policy with POLICY-ARN exists in AWS, nil otherwise.
+Uses \"aws iam get-policy\" and checks the exit code."
+  (let ((cmd (format "aws iam get-policy --policy-arn %s%s"
+                     (shell-quote-argument policy-arn)
+                     (org-aws-iam-role--cli-profile-arg))))
+    (eq 0 (call-process-shell-command cmd nil nil nil))))
+
+(defun org-aws-iam-role--babel-cmd-attach-managed-policy (role-name policy-arn)
+  "Return the AWS CLI command to attach a managed policy.
+ROLE-NAME is the IAM role name. POLICY-ARN is the ARN of the managed policy."
+  (format "aws iam attach-role-policy --role-name %s --policy-arn %s%s"
+          (shell-quote-argument role-name)
+          (shell-quote-argument policy-arn)
+          (org-aws-iam-role--cli-profile-arg)))
+
+(defun org-aws-iam-role--babel-cmd-create-policy (policy-name policy-document &optional path tags)
   "Return the AWS CLI command to create a customer-managed policy.
 POLICY-NAME is the name of the new policy.
-POLICY-DOCUMENT is the policy JSON string."
-  (format "aws iam create-policy --policy-name %s --policy-document %s%s"
+POLICY-DOCUMENT is the policy JSON string.
+PATH is the optional IAM path (e.g., \"/service-role/\").
+TAGS is the optional string of tags (e.g. \"Key=Owner,Value=Me\")."
+  (format "aws iam create-policy --policy-name %s --policy-document %s%s%s%s"
           (shell-quote-argument policy-name)
           (shell-quote-argument policy-document)
+          (if (and path (not (string-empty-p path)))
+              (format " --path %s" (shell-quote-argument path))
+            "")
+          (if (and tags (not (string-empty-p tags)))
+              (format " --tags %s" tags) ;; We assume user provides format "Key=K,Value=V ..."
+            "")
           (org-aws-iam-role--cli-profile-arg)))
 
+(defun org-aws-iam-role--tag-resource (role-name policy-arn policy-type tags)
+  "Apply TAGS to an existing resource based on POLICY-TYPE.
+If POLICY-TYPE is \='trust-policy', tags the IAM Role ROLE-NAME.
+If POLICY-TYPE is \='customer-managed', tags the IAM Policy POLICY-ARN.
+Asks for user confirmation before applying.
+Returns nil if no tagging was performed or user aborted."
+  (when (and tags (not (string-empty-p tags)))
+    (cond
+     ;; Case 1: Tag the Role (Context is Trust Policy)
+     ((eq policy-type 'trust-policy)
+      (let ((cmd (format "aws iam tag-role --role-name %s --tags %s%s"
+                         (shell-quote-argument role-name)
+                         tags
+                         (org-aws-iam-role--cli-profile-arg)))
+            (prompt (format "Apply tags '%s' to Role '%s'" tags role-name)))
+        (if (y-or-n-p (format "%s? " prompt))
+            (progn
+              (message "Executing: Tag Role...")
+              (shell-command-to-string cmd))
+          (message "Tagging aborted by user."))))
+
+     ;; Case 2: Tag the Managed Policy
+     ((eq policy-type 'customer-managed)
+      (let ((cmd (format "aws iam tag-policy --policy-arn %s --tags %s%s"
+                         (shell-quote-argument policy-arn)
+                         tags
+                         (org-aws-iam-role--cli-profile-arg)))
+            (prompt (format "Apply tags '%s' to Policy '%s'" tags policy-arn)))
+        (if (y-or-n-p (format "%s? " prompt))
+            (progn
+              (message "Executing: Tag Policy...")
+              (shell-command-to-string cmd))
+          (message "Tagging aborted by user."))))
+     
+     ;; Case 3: Inline
+     ((eq policy-type 'inline)
+      (message "Warning: Inline policies cannot be tagged directly. Tag the Role (Trust Policy) instead.")))))
+
 (defun org-aws-iam-role--babel-cmd-delete-inline-policy (role-name policy-name)
   "Return the AWS CLI command to delete an inline policy.
 ROLE-NAME is the IAM role name. POLICY-NAME is the inline policy name."
@@ -916,14 +976,42 @@ Argument POLICY-ARN is the ARN of the policy.
 Argument POLICY-TYPE is the type of the policy."
   (let (cmd action-desc)
     (cond
+     ;; CASE 1: Inline Policy
      ((eq policy-type 'inline)
       (setq action-desc (format "Permanently delete inline policy '%s' from role '%s'" policy-name role-name))
-      (setq cmd (org-aws-iam-role--babel-cmd-delete-inline-policy role-name policy-name)))
+      (setq cmd (org-aws-iam-role--babel-cmd-delete-inline-policy role-name policy-name))
+      (org-aws-iam-role--babel-confirm-and-run cmd action-desc))
+
+     ;; CASE 2: Customer Managed Policy
      ((eq policy-type 'customer-managed)
-      (setq action-desc (format "Permanently delete managed policy '%s'? (This will fail if it's still attached to any entity)" policy-arn))
-      (setq cmd (org-aws-iam-role--babel-cmd-delete-policy policy-arn)))
-     (t (user-error "Deletion is only supported for 'inline' and 'customer-managed' policies")))
-    (org-aws-iam-role--babel-confirm-and-run cmd action-desc)))
+      (let ((versions (org-aws-iam-role--get-all-non-default-versions policy-arn))
+            (delete-policy-cmd (org-aws-iam-role--babel-cmd-delete-policy policy-arn)))
+        
+        (if versions
+            ;; Sub-case: Multiple versions exist
+            (let ((count (length versions)))
+              (if (y-or-n-p (format "Policy has %d non-default versions. Delete ALL versions and the policy '%s'? " count policy-arn))
+                  (progn
+                    ;; 1. Delete all non-default versions
+                    (dolist (ver versions)
+                      (message "Deleting version %s..." ver)
+                      (let ((del-ver-cmd (format "aws iam delete-policy-version --policy-arn %s --version-id %s%s"
+                                                 (shell-quote-argument policy-arn)
+                                                 ver
+                                                 (org-aws-iam-role--cli-profile-arg))))
+                        (shell-command-to-string del-ver-cmd)))
+                    
+                    ;; 2. Delete the policy itself
+                    (message "Deleting policy %s..." policy-arn)
+                    (let ((result (string-trim (shell-command-to-string delete-policy-cmd))))
+                      (if (string-empty-p result) "Success! Policy and all versions deleted." result)))
+                (user-error "Aborted by user")))
+          
+          ;; Sub-case: No extra versions, just delete the policy
+          (setq action-desc (format "Permanently delete managed policy '%s'? (This will fail if it's still attached to any entity)" policy-arn))
+          (org-aws-iam-role--babel-confirm-and-run delete-policy-cmd action-desc))))
+
+     (t (user-error "Deletion is only supported for 'inline' and 'customer-managed' policies")))))
 
 (defun org-aws-iam-role--babel-handle-detach (role-name policy-name policy-arn policy-type)
   "Handle the :detach action for `aws-iam' babel blocks.
@@ -937,17 +1025,55 @@ Argument POLICY-TYPE is the type of the policy."
         (cmd (org-aws-iam-role--babel-cmd-detach-managed-policy role-name policy-arn)))
     (org-aws-iam-role--babel-confirm-and-run cmd action-desc)))
 
-(defun org-aws-iam-role--babel-handle-create (policy-name policy-type body)
-  "Handle the :create action for `aws-iam' babel blocks.
-Argument POLICY-NAME is the name of the new policy.
-Argument POLICY-TYPE is the type of the policy.
-Argument BODY is the JSON content of the policy."
+(defun org-aws-iam-role--babel-handle-create (role-name policy-name policy-type body &optional path tags)
+  "Handle the creation of a \='customer-managed' policy and optional attachment.
+If ROLE-NAME is provided, the new policy is attached to it immediately.
+ARGUMENTS: ROLE-NAME, POLICY-NAME, POLICY-TYPE, BODY, PATH, TAGS."
   (if (eq policy-type 'customer-managed)
       (let* ((json-string (json-encode (json-read-from-string body)))
-             (action-desc (format "Create new customer managed policy '%s'" policy-name))
-             (cmd (org-aws-iam-role--babel-cmd-create-policy policy-name json-string)))
-        (org-aws-iam-role--babel-confirm-and-run cmd action-desc))
-    (user-error "The :CREATE flag is only for 'customer-managed' policies. For inline policies, execute without it")))
+             (create-cmd (org-aws-iam-role--babel-cmd-create-policy policy-name json-string path tags))
+             (prompt (if role-name
+                         (format "Create policy '%s' (path: %s, tags: %s) AND attach to role '%s'"
+                                 policy-name (or path "/") (or tags "none") role-name)
+                       (format "Create new customer managed policy '%s' (path: %s, tags: %s)"
+                               policy-name (or path "/") (or tags "none")))))
+        
+        (if (y-or-n-p (format "%s? " prompt))
+            (progn
+              (message "Executing: Create Policy...")
+              ;; Capture the output and trim whitespace
+              (let ((output (string-trim (shell-command-to-string create-cmd))))
+                
+                ;; 1. Check for known CLI text errors (AWS errors often start with "An error...")
+                (when (string-prefix-p "An error" output)
+                  (user-error "AWS CLI Error: %s" output))
+
+                ;; 2. Attempt to parse JSON safely
+                (condition-case _err
+                    (let* ((parsed (json-parse-string output :object-type 'alist))
+                           (new-policy (alist-get 'Policy parsed))
+                           (new-arn (alist-get 'Arn new-policy)))
+                      
+                      (unless new-arn
+                        (error "JSON parsed, but no ARN found. Full Output: %s" output))
+
+                      (if role-name
+                          (progn
+                            (message "Executing: Attach Policy to Role...")
+                            (let* ((attach-cmd (org-aws-iam-role--babel-cmd-attach-managed-policy role-name new-arn))
+                                   (attach-output (string-trim (shell-command-to-string attach-cmd))))
+                              (if (not (string-empty-p attach-output))
+                                  (user-error "Created policy '%s' (ARN: %s), BUT attachment failed: %s"
+                                              policy-name new-arn attach-output)
+                                (format "Success! Created policy '%s' and attached it to '%s'.\nARN: %s"
+                                        policy-name role-name new-arn))))
+                        (format "Success! Created policy '%s'.\nARN: %s" policy-name new-arn)))
+                  
+                  ;; 3. Catch JSON parsing errors (e.g. other non-JSON text returned)
+                  (json-parse-error
+                   (user-error "Failed to parse AWS response. The command likely failed. Output: %s" output)))))
+          (user-error "Aborted by user")))
+    (user-error "The create action is only for 'customer-managed' policies")))
 
 (defun org-aws-iam-role--babel-handle-update (role-name policy-name policy-arn policy-type body)
   "Handle the default update action for `aws-iam' babel blocks.
@@ -956,48 +1082,167 @@ Argument POLICY-NAME is the name of the policy.
 Argument POLICY-ARN is the ARN of the policy.
 Argument POLICY-TYPE is the type of the policy.
 Argument BODY is the JSON content of the policy."
+  (cond
+   ;; 1. Trust Policies (Use standard confirm-and-run)
+   ((eq policy-type 'trust-policy)
+    (let* ((json-string (json-encode (json-read-from-string body)))
+           (action-desc (format "Update Trust Policy for role '%s'" role-name))
+           (cmd (org-aws-iam-role--babel-cmd-for-trust-policy role-name json-string)))
+      (org-aws-iam-role--babel-confirm-and-run cmd action-desc)))
+
+   ;; 2. Inline Policies (Use standard confirm-and-run)
+   ((eq policy-type 'inline)
+    (let* ((json-string (json-encode (json-read-from-string body)))
+           (action-desc (format "Update inline policy '%s' for role '%s'" policy-name role-name))
+           (cmd (org-aws-iam-role--babel-cmd-for-inline-policy role-name policy-name json-string)))
+      (org-aws-iam-role--babel-confirm-and-run cmd action-desc)))
+
+   ;; 3. Managed Policies (Use new retry logic)
+   ((or (eq policy-type 'customer-managed)
+        (eq policy-type 'aws-managed)
+        (eq policy-type 'permissions-boundary))
+    (org-aws-iam-role--update-managed-policy-with-retry policy-name policy-arn body))
+
+   (t (user-error "Unsupported policy type for modification: %s" policy-type))))
+
+(defun org-aws-iam-role--get-all-non-default-versions (policy-arn)
+  "Return a list of VersionIds for all non-default versions of POLICY-ARN."
+  (let* ((cmd (format "aws iam list-policy-versions --policy-arn %s --output json%s"
+                      (shell-quote-argument policy-arn)
+                      (org-aws-iam-role--cli-profile-arg)))
+         (json (shell-command-to-string cmd))
+         (parsed (condition-case nil
+                     (json-parse-string json :object-type 'alist :array-type 'list)
+                   (error nil)))
+         (versions (alist-get 'Versions parsed)))
+    (when versions
+      ;; Return only VersionIds where IsDefaultVersion is NOT true
+      (mapcar (lambda (v) (alist-get 'VersionId v))
+              (cl-remove-if (lambda (v) (eq t (alist-get 'IsDefaultVersion v))) versions)))))
+
+(defun org-aws-iam-role--get-oldest-non-default-version (policy-arn)
+  "Return the VersionId of the oldest non-default version for POLICY-ARN."
+  (let* ((cmd (format "aws iam list-policy-versions --policy-arn %s --output json%s"
+                      (shell-quote-argument policy-arn)
+                      (org-aws-iam-role--cli-profile-arg)))
+         (json (shell-command-to-string cmd))
+         (parsed (json-parse-string json :object-type 'alist :array-type 'list))
+         (versions (alist-get 'Versions parsed))
+         (non-defaults (cl-remove-if (lambda (v) (eq t (alist-get 'IsDefaultVersion v))) versions)))
+    (when non-defaults
+      (let ((sorted (sort non-defaults
+                          (lambda (a b)
+                            (string< (alist-get 'CreateDate a)
+                                     (alist-get 'CreateDate b))))))
+        (alist-get 'VersionId (car sorted))))))
+
+(defun org-aws-iam-role--update-managed-policy-with-retry (policy-name policy-arn body)
+  "Update managed policy POLICY-NAME (POLICY-ARN) with BODY.
+Handles version limits by offering to delete the oldest version."
   (let* ((json-string (json-encode (json-read-from-string body)))
-         (action-desc (format "Update %s for role '%s'"
-                              (if (eq policy-type 'trust-policy) "Trust Policy" (format "policy '%s'" policy-name))
-                              role-name))
-         (cmd (cond
-               ((eq policy-type 'trust-policy)
-                (org-aws-iam-role--babel-cmd-for-trust-policy role-name json-string))
-               ((eq policy-type 'inline)
-                (org-aws-iam-role--babel-cmd-for-inline-policy role-name policy-name json-string))
-               ((or (eq policy-type 'customer-managed)
-                    (eq policy-type 'aws-managed)
-                    (eq policy-type 'permissions-boundary))
-                (org-aws-iam-role--babel-cmd-for-managed-policy policy-arn json-string))
-               (t (user-error "Unsupported policy type for modification: %s" policy-type)))))
-    (org-aws-iam-role--babel-confirm-and-run cmd action-desc)))
+         (cmd (org-aws-iam-role--babel-cmd-for-managed-policy policy-arn json-string))
+         (action-desc (format "Update managed policy '%s' (ARN: %s)" (or policy-name "unknown") policy-arn)))
+    
+    (if (y-or-n-p (format "%s? " action-desc))
+        (progn
+          (message "Executing: Update Policy...")
+          (let ((result (string-trim (shell-command-to-string cmd))))
+            ;; Check for LimitExceededException in the output
+            (if (and (not (string-empty-p result))
+                     (string-match-p "LimitExceeded" result))
+                
+                ;; limit reached logic
+                (let ((oldest-ver (org-aws-iam-role--get-oldest-non-default-version policy-arn)))
+                  (if (and oldest-ver
+                           (y-or-n-p (format "Version limit reached. Delete oldest version (%s) and retry?" oldest-ver)))
+                      (progn
+                        (message "Deleting version %s..." oldest-ver)
+                        (let ((del-cmd (format "aws iam delete-policy-version --policy-arn %s --version-id %s%s"
+                                               (shell-quote-argument policy-arn)
+                                               oldest-ver
+                                               (org-aws-iam-role--cli-profile-arg))))
+                          (shell-command-to-string del-cmd))
+                        (message "Retrying update...")
+                        ;; Retry the original update command
+                        (let ((retry-result (string-trim (shell-command-to-string cmd))))
+                          (if (string-empty-p retry-result) "Success!" retry-result)))
+                    
+                    ;; User said no to deletion, or we couldn't find a version to delete
+                    (user-error "Update failed (Limit Exceeded): %s" result)))
+              
+              ;; normal success/failure logic
+              (if (string-match-p "An error occurred" result)
+                  (user-error "Update failed: %s" result)
+                (if (string-empty-p result) "Success!" result)))))
+      (user-error "Aborted by user"))))
 
 (defun org-babel-execute:aws-iam (body params)
   "Execute an `aws-iam' source block.
 BODY with header PARAMS to manage IAM policies.
 PARAMS should include header arguments such as :ROLE-NAME, :POLICY-NAME,
-:ARN, and :POLICY-TYPE."
+:ARN, :POLICY-TYPE, :PATH, and :TAGS."
   (when buffer-read-only
     (user-error "Buffer is read-only. Press C-c C-e to enable edits and execution"))
 
   (let* ((role-name (cdr (assoc :role-name params)))
          (policy-name (cdr (assoc :policy-name params)))
          (policy-arn (cdr (assoc :arn params)))
+         (policy-path (cdr (assoc :path params)))
+         (tags (cdr (assoc :tags params)))
          (policy-type-str (cdr (assoc :policy-type params)))
          (policy-type (when policy-type-str (intern policy-type-str)))
-         (create-p (org-aws-iam-role--param-true-p (cdr (assoc :create params))))
          (delete-p (org-aws-iam-role--param-true-p (cdr (assoc :delete params))))
-         (detach-p (org-aws-iam-role--param-true-p (cdr (assoc :detach params)))))
+         (detach-p (org-aws-iam-role--param-true-p (cdr (assoc :detach params))))
+         (create-p nil))
+
+    ;; LOGIC FOR CUSTOMER-MANAGED POLICIES ("Upsert" Logic)
+    (when (eq policy-type 'customer-managed)
+      (unless policy-name
+        (user-error "Header argument :policy-name is required for customer-managed policies"))
+      
+      ;; 1. If ARN is missing OR path is provided, synthesize/recalculate ARN
+      (when (or policy-path
+                (not policy-arn)
+                (string-empty-p policy-arn))
+        (let ((account-id (org-aws-iam-role--get-account-id)))
+          (unless account-id (error "Could not fetch AWS Account ID to construct ARN"))
+          
+          (let* ((raw-path (if (and policy-path (not (string-empty-p policy-path))) policy-path "/"))
+                 (p-start (if (string-prefix-p "/" raw-path) raw-path (concat "/" raw-path)))
+                 (p-full (if (string-suffix-p "/" p-start) p-start (concat p-start "/")))
+                 (clean-path (substring p-full 1)))
+            
+            (setq policy-arn (format "arn:aws:iam::%s:policy/%s%s"
+                                     account-id clean-path policy-name)))))
+
+      ;; 2. Check existence
+      (if (org-aws-iam-role--policy-exists-p policy-arn)
+          (setq create-p nil)
+        (setq create-p t)))
 
     (unless (and (or role-name create-p) policy-type)
       (user-error "Missing required header arguments: :ROLE-NAME or :POLICY-TYPE"))
 
-    (cond
-     (delete-p (org-aws-iam-role--babel-handle-delete role-name policy-name policy-arn policy-type))
-     (detach-p (org-aws-iam-role--babel-handle-detach role-name policy-name policy-arn policy-type))
-     (create-p (org-aws-iam-role--babel-handle-create policy-name policy-type body))
-     (t (org-aws-iam-role--babel-handle-update role-name policy-name policy-arn policy-type body)))))
-
+    ;; We collect output messages in a list to support multiple sequential actions.
+    (let ((results '()))
+      ;; 1. TAGGING (Priority 1)
+      ;; Only if we are NOT destroying the resource.
+      (when (and (not create-p) tags (not delete-p) (not detach-p))
+        (org-aws-iam-role--tag-resource role-name policy-arn policy-type tags))
+      ;; 2. DETACH (Priority 2)
+      (when detach-p
+        (if (eq policy-type 'inline)
+            (message "Skipping detach for inline policy (implicit in delete).")
+          (push (org-aws-iam-role--babel-handle-detach role-name policy-name policy-arn policy-type) results)))
+      ;; 3. DELETE (Priority 3)
+      (when delete-p
+        (push (org-aws-iam-role--babel-handle-delete role-name policy-name policy-arn policy-type) results))
+      ;; 4. CREATE / UPDATE (Priority 4)
+      (unless (or detach-p delete-p)
+        (if create-p
+            (push (org-aws-iam-role--babel-handle-create role-name policy-name policy-type body policy-path tags) results)
+          (push (org-aws-iam-role--babel-handle-update role-name policy-name policy-arn policy-type body) results)))
+      (mapconcat #'identity (nreverse results) "\n"))))
 
 ;;;;; unified json start ;;;;;
 
@@ -1014,6 +1259,14 @@ Removes all non-alphanumeric characters."
   (when s
     (replace-regexp-in-string "[^a-zA-Z0-9]" "" s)))
 
+(defun org-aws-iam-role--parse-modified-date (date-string)
+  "Parse a DATE-STRING into an Emacs time list.
+Return nil if the string is nil, \"nil\", or invalid."
+  (when (and date-string (not (string-empty-p date-string)) (not (string= "nil" date-string)))
+    (condition-case nil
+        (parse-time-string date-string)
+      (error nil))))
+
 (defun org-aws-iam-role--extract-all-permission-statements ()
   "Parse the current buffer to find and extract all permission policy statements.
 This function uses a state machine to iterate through headlines,
@@ -1042,7 +1295,12 @@ processed statement alists."
               (when src-block
                 (let* ((json-string (org-element-property :value src-block))
                        (policy-data (json-parse-string json-string :object-type 'alist :array-type 'list))
-                       (statements (alist-get 'Statement policy-data)))
+                       (raw-statements (alist-get 'Statement policy-data))
+                       (statements (if (and (consp raw-statements)
+                                            (consp (car raw-statements))
+                                            (atom (caar raw-statements)))
+                                       (list raw-statements)
+                                     raw-statements)))
                   (when statements
                     (dolist (stmt statements)
                       (let* ((stmt-no-sid (cl-remove 'Sid stmt :key #'car))
@@ -1085,6 +1343,53 @@ to extract policy statements and display them in a new buffer."
       (let ((role-name (org-aws-iam-role--get-role-name-from-buffer)))
         (org-aws-iam-role--create-and-show-json-buffer all-statements role-name)))))
 
+;;;###autoload
+(defun org-aws-iam-role-get-last-modified ()
+  "Find the most recent :Created: or :Updated: timestamp in the buffer.
+This parses all property drawers to find the latest date,
+which represents the last known modification time of the role or
+one of its policies."
+  (interactive)
+  (unless (derived-mode-p 'org-mode)
+    (user-error "This command must be run from an Org mode buffer"))
+  (let* ((tree (org-element-parse-buffer))
+         (all-dates (org-aws-iam-role--collect-dates-from-buffer tree)))
+    (org-aws-iam-role--find-latest-date all-dates)))
+
+(defun org-aws-iam-role--collect-dates-from-buffer (tree)
+  "Collect all :Created: and :Updated: dates from the parse TREE."
+  (let ((all-dates '()))
+    (org-element-map tree 'property-drawer
+      (lambda (drawer)
+        (org-element-map (org-element-contents drawer) 'node-property
+          (lambda (prop)
+            (let ((key (org-element-property :key prop))
+                  (value (org-element-property :value prop)))
+              (when (and value (or (string= key "Created") (string= key "Updated")))
+                (let ((parsed-date (org-aws-iam-role--parse-modified-date value)))
+                  (when parsed-date
+                    (push parsed-date all-dates)))))))))
+    all-dates))
+
+(defun org-aws-iam-role--find-latest-date (all-dates)
+  "Find the latest date in ALL-DATES and return it as a formatted string."
+  (if (null all-dates)
+      (message "No modified dates found in this buffer.")
+    (let* ((latest-time-DECODED (car all-dates))
+           (time-val-current nil)
+           (time-val-latest nil)
+           (is-greater nil)
+           (final-time-VALUE nil))
+      (dolist (current-date-DECODED (cdr all-dates))
+        (setq time-val-current (apply #'encode-time current-date-DECODED))
+        (setq time-val-latest (apply #'encode-time latest-time-DECODED))
+        (setq is-greater (time-less-p time-val-latest time-val-current))
+        (when is-greater
+          (setq latest-time-DECODED current-date-DECODED)))
+      (setq final-time-VALUE (apply #'encode-time latest-time-DECODED))
+      (let ((latest-date-string (format-time-string "%FT%T%z" final-time-VALUE)))
+        (message "Last modification date found: %s" latest-date-string)))))
+
 ;;;;; Last Accessed Details ;;;;;
 
 (defun org-aws-iam-role--show-last-accessed-json (json-string role-arn)
